import os

import pandas as pd
from pandas.testing import assert_frame_equal
from pydantic import BaseModel

from pytest_dbt_duckdb.connector import DuckConnector
from pytest_dbt_duckdb.dbt_executor import DbtExecutor
from pytest_dbt_duckdb.sql_methods import create_schema, create_table


class DbtTestNode(BaseModel):
    schema: str  # type: ignore
    table: str
    path: str

    @property
    def model(self) -> str:
        return f"{self.schema}.{self.table}"


class DbtValidator:
    def __init__(
        self, connector: DuckConnector, executor: DbtExecutor, resources_folder: str, debug_output: bool = False
    ) -> None:
        self.connector = connector
        self.executor = executor
        self.resources_folder = resources_folder

        # Initialize validator
        self.nodes = self.executor.parse_project()
        self.debug_output = debug_output

    def dbt_create_model(
        self,
        node: DbtTestNode,
        include_columns: list[str] | None = None,
        table_alias: str | None = None,
    ) -> None:
        source = self.nodes[node.model]
        create_schema(self.connector, source.schema)

        columns = source.columns
        if include_columns:
            columns = {column: columns[column] for column in include_columns}

        create_table(self.connector, source.schema, table_alias or source.identifier, list(columns.values()))

    def dbt_insert_model(
        self, node: DbtTestNode, include_columns: list[str] | None = None, table_alias: str | None = None
    ) -> None:
        file = os.path.join(self.resources_folder, node.path)

        columns = None
        if dbt_node := self.nodes.get(node.model):
            columns = {field.name: field.data_type for field in dbt_node.columns.values() if field.data_type}
            if include_columns:
                columns = {column: columns[column] for column in include_columns}

            for column, data_type in columns.items():
                if data_type and data_type.startswith("MAP"):
                    columns[column] = "JSON"

        self.connector.insert_data(
            table=table_alias or node.model,
            data_path=file,
            columns=columns,
        )

    def extract_node_columns(self, node: DbtTestNode) -> list[str]:
        file = os.path.join(self.resources_folder, node.path)
        self.connector.create_tmp_table(table=f"temp_{node.table}", data_path=file)
        return self.connector.get_table_columns(f"temp_{node.table}")

    def dbt_load_node(self, node: DbtTestNode) -> None:
        file_columns = self.extract_node_columns(node=node)

        try:
            self.dbt_create_model(node=node, include_columns=file_columns)
            self.dbt_insert_model(node=node, include_columns=file_columns)
        except Exception as error:
            print(f"Error  {error} at node {node}")
            raise error

    def dbt_load_nodes(self, nodes: list[DbtTestNode]) -> None:
        for node in nodes:
            self.dbt_load_node(node)

    def execute_dbt(self, seed: str | None = None, build: str | None = None, quiet: bool = True) -> None:
        quiet_param = "-q" if quiet else ""

        assert seed is not None or build is not None, "seed or build must be defined"

        if seed:
            seeds_res = self.executor.execute(command="seed", params=["--select", seed, quiet_param])
            self.executor.validate_execution(seeds_res)
        if build:
            build_res = self.executor.execute(command="build", params=["--select", build, quiet_param])
            self.executor.validate_execution(build_res)

    @staticmethod
    def display_df(node: DbtTestNode, df: pd.DataFrame, print_json: bool = False) -> None:
        pd.set_option("display.max_columns", None)
        pd.set_option("display.expand_frame_repr", True)
        pd.set_option("display.max_rows", None)
        pd.set_option("display.max_seq_items", None)
        pd.set_option("display.max_colwidth", 500)
        pd.set_option("expand_frame_repr", False)
        df = df.copy()  # Create a copy of the dataframe to avoid modifying the original
        for column in df.select_dtypes(include=["datetime"]):
            df[column] = df[column].astype(str)
        if print_json:
            df.to_csv(
                path_or_buf=f"/tmp/print_{node.table}.csv",
                index=False,
            )
            pass
        print(df)

    def dbt_validate_node(self, node: DbtTestNode) -> None:
        df = self.connector.fetch_table(schema=node.schema, table=node.table)

        expected_table = f"{node.table}_test"
        expected_model = f"{node.schema}.{expected_table}"

        fixture_columns = self.extract_node_columns(node=node)
        try:
            # Execute taking the dbt documentation as reference
            self.dbt_create_model(node=node, table_alias=expected_table, include_columns=fixture_columns)
            self.dbt_insert_model(node=node, table_alias=expected_model, include_columns=fixture_columns)
        except:  # noqa
            # Execute taking the dbt output as reference
            self.connector.clone_table(source=node.model, target=expected_model)
            self.dbt_insert_model(node=node, table_alias=expected_model)
        finally:
            expected_df = self.connector.fetch_table(schema=node.schema, table=expected_table)

            try:
                assert_frame_equal(df, expected_df, rtol=1e-5, atol=1e-8)
            except AssertionError as error:
                print(f"Error at node {node.table}")
                self.display_df(node=node, df=df, print_json=self.debug_output)
                self.display_df(node=node, df=expected_df)
                raise error

    def dbt_validate_nodes(self, nodes: list[DbtTestNode]) -> None:
        for node in nodes:
            self.dbt_validate_node(node=node)

    def validate(
        self,
        nodes_to_load: list[DbtTestNode],
        nodes_to_validate: list[DbtTestNode],
        seed: str | None = None,
        build: str | list[str] | None = None,
    ) -> None:
        # STEP 1: Populate Source Tables with CSV/JSON files
        self.dbt_load_nodes(nodes=nodes_to_load)

        # STEP 2: Execute [seed|run|test] jobs & fail if execution errors
        if build and isinstance(build, list):
            build = " ".join(build)
        self.execute_dbt(seed=seed, build=str(build), quiet=False)

        # STEP 3: Validate Output Tables versus CSV files
        self.dbt_validate_nodes(nodes=nodes_to_validate)



import pytest

from pytest_dbt_duckdb.plugin import DuckFixture, TestFixture, load_yaml_tests
from tests import dbt_project_dir, resources_folder

yaml_data = list(load_yaml_tests(resources_folder))


@pytest.mark.parametrize("fixture", yaml_data, ids=[x.id for x in yaml_data])
def test_dbt_scenarios(fixture: TestFixture, duckdb_fixture: DuckFixture) -> None:
    duckdb_fixture.execute_dbt(
        nodes_to_load=fixture.given,
        seed=fixture.seed,
        build=fixture.build,
        nodes_to_validate=fixture.then,
        resources_folder=resources_folder,
        dbt_project_dir=dbt_project_dir,
    )



from datetime import datetime, timedelta

from dateutil.relativedelta import relativedelta


def to_char(dt: datetime, fmt: str) -> str | None:
    if not dt:
        return None
    fmt = fmt.replace("YYYY", "%Y").replace("MM", "%m").replace("DD", "%d")
    return dt.strftime(fmt)


def date_add(unit: str, value: int, timestamp: datetime) -> datetime:
    unit = unit.lower()

    if unit == "second":
        return timestamp + timedelta(seconds=value)
    elif unit == "minute":
        return timestamp + timedelta(minutes=value)
    elif unit == "hour":
        return timestamp + timedelta(hours=value)
    elif unit == "day" or unit == "days":
        return timestamp + timedelta(days=value)
    elif unit == "week":
        return timestamp + timedelta(weeks=value)
    elif unit == "month":
        return timestamp + relativedelta(months=value)
    elif unit == "year":
        return timestamp + relativedelta(years=value)
    else:
        raise ValueError(f"Unsupported unit: {unit}")


def iff() -> str:
    return """
        CREATE MACRO IF NOT EXISTS iff(condition, true_value, false_value) AS (
            CASE
                WHEN condition THEN true_value
                ELSE false_value
            END
        );
    """


def bitand() -> str:
    return "CREATE MACRO IF NOT EXISTS BITAND(a, b) AS (a & b);"


def array_size() -> str:
    return """
        CREATE MACRO IF NOT EXISTS array_size(arr) AS (
            CASE
                WHEN arr IS NULL THEN NULL
                ELSE ARRAY_LENGTH(arr)
            END
        );
    """


def array_union_agg() -> str:
    return """
        CREATE MACRO IF NOT EXISTS array_union_agg(field) AS (
            list_distinct(flatten(list(field)))
        );
    """


def to_decimal() -> str:
    return """
        CREATE MACRO IF NOT EXISTS to_decimal(expr, precision, scale) AS (
            CAST(expr as DECIMAL(12,5))
        );
    """


def div0() -> str:
    return """
        CREATE MACRO IF NOT EXISTS div0(numerator, denominator) AS (
            CASE
                WHEN denominator = 0 THEN 0
                ELSE numerator / denominator
            END
        );
    """


def current_timestamp() -> str:
    return """
        CREATE MACRO IF NOT EXISTS current_timestamp() AS (
            current_localtimestamp()
        );
    """



from __future__ import annotations

import json
from typing import Any, Callable

import numpy as np
import pandas as pd
from absl import logging
from duckdb import DuckDBPyConnection
from duckdb.typing import DATE, INTEGER, VARCHAR, DuckDBPyType
from pydantic import BaseModel, ConfigDict

from pytest_dbt_duckdb.snowflake_functions import (
    array_size,
    array_union_agg,
    bitand,
    current_timestamp,
    date_add,
    div0,
    iff,
    to_char,
    to_decimal,
)


class DuckFunction(BaseModel):
    name: str
    function: Callable
    parameters: list[DuckDBPyType | Any] | None = None
    return_type: DuckDBPyType | Any | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class ExtraFunctions(BaseModel):
    macros: list[str] | None = None
    functions: list[DuckFunction] | None = None


class DuckConnector:
    def __init__(self, conn: DuckDBPyConnection, extra_functions: ExtraFunctions | None) -> None:
        self.conn = conn
        self.register_snowflake_functions()

        if extra_functions:
            for macro in extra_functions.macros or []:
                self.execute(query=macro)
            for fn in extra_functions.functions or []:
                self.conn.create_function(
                    name=fn.name,
                    function=fn.function,
                    parameters=fn.parameters,
                    return_type=fn.return_type,
                )

    def execute(self, query: str, parameters: dict[str, Any] | None = None) -> DuckDBPyConnection:
        logging.info(f"Running query = {query}")
        return self.conn.execute(query, parameters=parameters)

    def clone_table(self, source: str, target: str) -> DuckDBPyConnection:
        return self.execute(f"CREATE OR REPLACE TABLE {target} AS SELECT * FROM {source} WHERE 1=0;")

    def insert_data(self, table: str, data_path: str, columns: dict[str, str] | None = None) -> DuckDBPyConnection:
        if data_path.endswith(".csv"):
            if columns:
                return self.execute(f"""
                    INSERT INTO {table}
                    SELECT * FROM read_csv(
                        '{data_path}',
                        header = true,
                        allow_quoted_nulls = false,
                        auto_detect = true,
                        columns={columns}
                    )
                """)
            return self.execute(f"INSERT INTO {table} SELECT * FROM read_csv_auto('{data_path}')")
        elif data_path.endswith(".json"):
            if columns:
                return self.execute(f"""
                    INSERT INTO {table}
                    SELECT * FROM read_json(
                        '{data_path}',
                        format='array',
                        columns={columns}
                    )
                """)
            return self.execute(f"INSERT INTO {table} SELECT * FROM read_json_auto('{data_path}')")
        else:
            raise RuntimeError(f"File {data_path} not supported")

    def fetch_data(self, query: str, parameters: dict[str, Any] | None = None) -> list[dict]:
        return self.execute(query, parameters=parameters).arrow().to_pylist()

    @staticmethod
    def sort_df(df: pd.DataFrame) -> pd.DataFrame:
        def make_hashable(value: Any) -> Any:
            if isinstance(value, np.ndarray):
                return make_hashable(value.tolist())
            elif isinstance(value, (list, dict)):
                return json.dumps(value, sort_keys=True)
            return value

        df = df.map(make_hashable)
        df = df.reindex(sorted(df.columns), axis=1)
        return df.sort_values(by=list(df.columns), axis=0).reset_index(drop=True)

    def fetch_sorted_df(self, query: str, parameters: dict[str, Any] | None = None) -> pd.DataFrame:
        df = self.execute(query, parameters=parameters).fetch_df()
        return self.sort_df(df)

    def fetch_table(self, schema: str, table: str) -> pd.DataFrame:
        return self.fetch_sorted_df(f"SELECT * FROM {schema}.{table}")

    def describe_table(self, schema: str, table: str) -> pd.DataFrame:
        return self.fetch_sorted_df(f"SHOW {schema}.{table}")

    def get_table_columns(self, table: str) -> list[str]:
        columns = self.execute(f"SELECT name FROM pragma_table_info('{table}')").arrow().to_pydict()["name"]
        return columns

    def create_tmp_table(self, table: str, data_path: str) -> DuckDBPyConnection:
        if data_path.endswith(".csv"):
            return self.execute(f"CREATE TEMP TABLE {table} AS SELECT * FROM read_csv_auto('{data_path}');")
        elif data_path.endswith(".json"):
            return self.execute(f"CREATE TEMP TABLE {table} AS SELECT * FROM read_json_auto('{data_path}');")
        else:
            raise RuntimeError(f"File {data_path} not supported")

    def commit(self) -> None:
        self.conn.commit()

    def rollback(self) -> None:
        self.conn.rollback()

    def __exit__(
        self,
        exception_type: str,
        exception_value: str,
        traceback: str,
    ) -> None:
        self.conn.close()
        logging.info("Closing connection...")

    def __del__(self) -> None:
        self.conn.close()
        logging.info("Closing connection...")

    def __enter__(self) -> DuckConnector:
        return self

    def register_snowflake_functions(self) -> None:
        self.execute(query=iff())
        self.execute(query=bitand())

        self.execute(query=array_size())
        self.execute(query=array_union_agg())
        self.execute(query=div0())
        self.execute(query=to_decimal())
        self.execute(query=current_timestamp())
        self.conn.create_function(
            name="to_char",
            function=to_char,
            parameters=[DATE, VARCHAR],
            return_type=VARCHAR,
        )
        self.conn.create_function(
            name="dateadd",
            function=date_add,
            parameters=[VARCHAR, INTEGER, DATE],
            return_type=DATE,
        )



import json

from absl import logging
from dbt.artifacts.schemas.results import TestStatus
from dbt.cli.main import dbtRunner, dbtRunnerResult
from dbt.contracts.graph.nodes import ModelNode, SourceDefinition


class DbtExecException(Exception):
    pass


class DbtExecutor:
    def __init__(self, dbt_project_dir: str, profiles_dir: str) -> None:
        self.dbt_project_dir = dbt_project_dir
        self.profiles_dir = profiles_dir

    def execute(self, command: str, params: list | None = None) -> dbtRunnerResult:
        dbt = dbtRunner()
        params = params or []
        invoke_command = [
            command,
            "--project-dir",
            self.dbt_project_dir,
            "--profiles-dir",
            self.profiles_dir,
            "--vars",
            json.dumps({"elementary_enabled": False}),
        ]
        dbt_command = list(filter(lambda x: x, invoke_command + params))
        logging.info(f"DBT execute {dbt_command}")
        return dbt.invoke(dbt_command)

    def parse_project(self) -> dict[str, SourceDefinition | ModelNode]:
        res: dbtRunnerResult = self.execute(command="parse")

        result_sources: list[SourceDefinition] = res.result.sources.values()  # type: ignore
        sources = [source for source in result_sources if source.columns]

        result_models: list[ModelNode] = res.result.nodes.values()  # type: ignore
        models = [model for model in result_models if model.columns]

        nodes = sources + models
        return {f"{node.schema}.{node.identifier}": node for node in nodes}

    @staticmethod
    def validate_execution(res: dbtRunnerResult) -> None:
        errors = (
            result
            for result in res.result.results  # type: ignore
            if result and result.status in [TestStatus.Fail, TestStatus.Error]  # type: ignore
        )
        for error in errors:
            raise DbtExecException(f"Issue in {error.node.name} - {error.message}")



import os
import tempfile
from typing import Iterable

import duckdb
import pytest
from duckdb import DuckDBPyConnection
from pydantic import BaseModel, ConfigDict
from pydantic_settings import BaseSettings, SettingsConfigDict
from ruamel.yaml import YAML

from pytest_dbt_duckdb.connector import DuckConnector, ExtraFunctions
from pytest_dbt_duckdb.dbt_executor import DbtExecutor
from pytest_dbt_duckdb.dbt_validator import DbtTestNode, DbtValidator


class TestFixture(BaseModel):
    id: str
    given: list[DbtTestNode]
    build: str | list[str] | None = None
    seed: str | None = None
    then: list[DbtTestNode]


class PyDuckSettings(BaseSettings):
    temp_dir: str
    database_name: str = "dbt_duck"
    debug_output: bool = False
    model_config = SettingsConfigDict(env_prefix="dbt_")

    @property
    def db_file_path(self) -> str:
        return os.path.join(self.temp_dir, self.database_file)

    @property
    def database_file(self) -> str:
        return f"{self.database_name}.duckdb"


class DuckFixture(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    conn: DuckDBPyConnection
    settings: PyDuckSettings

    def execute_dbt(
        self,
        dbt_project_dir: str,
        resources_folder: str,
        nodes_to_load: list[DbtTestNode],
        nodes_to_validate: list[DbtTestNode],
        seed: str | None = None,
        build: str | list[str] | None = None,
        extra_functions: ExtraFunctions | None = None,
    ) -> None:
        connector = DuckConnector(conn=self.conn, extra_functions=extra_functions)
        os.environ["DBT_DUCKDB_PATH"] = self.settings.db_file_path
        os.environ["DBT_DUCKDB_DATABASE"] = self.settings.database_name

        executor = DbtExecutor(dbt_project_dir=dbt_project_dir, profiles_dir=resources_folder)
        validator = DbtValidator(connector=connector, executor=executor, resources_folder=resources_folder)
        validator.validate(nodes_to_load=nodes_to_load, nodes_to_validate=nodes_to_validate, seed=seed, build=build)


def load_yaml_test(file_path: str, yaml: YAML = YAML(typ="safe", pure=True)) -> Iterable[TestFixture]:
    with open(file_path, "r") as file:
        tests: list[dict] = yaml.load(file)["tests"]
        for test_fixture in tests:
            yield TestFixture(**test_fixture)


def load_yaml_tests(directory: str) -> Iterable[TestFixture]:
    yaml = YAML(typ="safe", pure=True)
    for filename in os.listdir(directory):
        if filename.startswith("test") & (filename.endswith(".yaml") or filename.endswith(".yml")):
            file_path = os.path.join(directory, filename)
            yield from load_yaml_test(file_path=file_path, yaml=yaml)


@pytest.fixture(scope="function")
def duckdb_fixture() -> Iterable[DuckFixture]:
    temp_dir = tempfile.gettempdir()
    settings = PyDuckSettings(temp_dir=str(temp_dir))

    conn = duckdb.connect(settings.db_file_path)
    try:
        yield DuckFixture(conn=conn, settings=settings)
    finally:
        conn.close()

    # Remove the file after the test
    if os.path.exists(settings.db_file_path):
        os.remove(settings.db_file_path)



from dbt.contracts.graph.nodes import ColumnInfo

from pytest_dbt_duckdb.connector import DuckConnector


def create_schema(connector: DuckConnector, schema: str) -> None:
    connector.execute(f"CREATE SCHEMA IF NOT EXISTS {schema};")


def create_table(connector: DuckConnector, schema: str, table: str, columns: list[ColumnInfo]) -> None:
    for column in columns:
        if column.data_type and column.data_type.startswith("MAP"):
            column.data_type = "JSON"

    columns_sql = [f'"{column.name}" {column.data_type or "varchar"}' for column in columns]
    columns_str = ",".join(columns_sql)
    ddl = f"CREATE OR REPLACE TABLE {schema}.{table} ({columns_str})"
    connector.execute(ddl)
